@@ -78,7 +78,7 @@ BEGIN {
 	}) {
 		@ISA = qw(IO::Socket::INET);
 	}
-	$VERSION = '1.37';
+	$VERSION = '1.54';
 	$GLOBAL_CONTEXT_ARGS = {};
 
 	#Make $DEBUG another name for $Net::SSLeay::trace
@@ -121,12 +121,14 @@ BEGIN {
 		*{idn_to_ascii} = \&Net::IDN::Encode::domain_to_ascii;
 	} elsif ( eval { require Net::LibIDN }) {
 		*{idn_to_ascii} = \&Net::LibIDN::idn_to_ascii;
+	} elsif ( eval { require URI; URI->VERSION(1.50) }) {
+	        *{idn_to_ascii} = sub { URI->new("http://" . shift)->host }
 	} else {
 		# default: croak if we really got an unencoded international domain
 		*{idn_to_ascii} = sub {
 			my $domain = shift;
 			return $domain if $domain =~m{^[a-zA-Z0-9-_\.]+$};
-			croak "cannot handle international domains, please install Net::LibIDN or Net::IDN::Encode"
+			croak "cannot handle international domains, please install Net::LibIDN, Net::IDN::Encode or URI"
 		}
 	}
 }
@@ -285,7 +287,7 @@ sub configure_SSL {
 		$arg_hash->{SSL_verify_callback} = sub {
 			my ($ok,$ctx_store,$certname,$error,$cert) = @_;
 			$ok = $vcb->($ok,$ctx_store,$certname,$error,$cert) if $vcb;
-			$ok or return;
+			$ok or return 0;
 			my $depth = Net::SSLeay::X509_STORE_CTX_get_error_depth($ctx_store);
 			return $ok if $depth != 0;
 
@@ -716,14 +718,27 @@ sub getc {
 
 sub readline {
 	my $self = shift;
-	my $ssl = $self->_get_ssl_object || return;
 
-	if (wantarray) {
-		my ($buf,$err) = Net::SSLeay::ssl_read_all($ssl);
-		return $self->error( "SSL read error" ) if $err;
-		if ( !defined($/) ) {
-			return $buf;
-		} elsif ( ref($/) ) {
+	if ( not defined $/ or wantarray) {
+		# read all and split
+
+		my $buf = '';
+		my $was_blocking = $self->blocking(1);
+		while (1) {
+			my $rv = $self->sysread($buf,2**16,length($buf));
+			if ( ! defined $rv ) {
+				next if $!{EINTR} or $!{EAGAIN};
+				$self->blocking(0) if ! $was_blocking;
+				return;
+			} elsif ( ! $rv ) {
+				last
+			}
+		}
+		$self->blocking(0) if ! $was_blocking;
+
+		if ( ! defined $/ ) {
+			return $buf
+		} elsif ( ref($/)) {
 			my $size = ${$/};
 			die "bad value in ref \$/: $size" unless $size>0;
 			return $buf=~m{\G(.{1,$size})}g;
@@ -734,59 +749,109 @@ sub readline {
 		}
 	}
 
-	if ( !defined($/) ) {
-		my ($buf,$err) = Net::SSLeay::ssl_read_all($ssl);
-		return $self->error( "SSL read error" ) if $err;
-		return $buf;
-	} elsif ( ref($/) ) {
+	# read only one line
+	if ( ref($/) ) {
 		my $size = ${$/};
+		# read record of $size bytes
 		die "bad value in ref \$/: $size" unless $size>0;
-		my ($buf,$err) = Net::SSLeay::ssl_read_all($ssl,$size);
-		return $self->error( "SSL read error" ) if $err;
-		return $buf;
-	} elsif ( $/ ne '' ) {
-		my $line = Net::SSLeay::ssl_read_until($ssl,$/);
-		return $self->error( "SSL read error" ) if $line eq '';
-		return $line;
-	} else {
-		# $/ is ''
-		# ^.*?\n\n+, need peek to find all \n at the end
-		die "empty \$/ is not supported if I don't have peek"
-			if Net::SSLeay::OPENSSL_VERSION_NUMBER() < 0x0090601f;
-
-		# find first occurence of \n\n
 		my $buf = '';
-		my $eon = 0;
-		while (1) {
-			defined( Net::SSLeay::peek($ssl,1)) || last; # peek more, can block
-			my $pending = Net::SSLeay::pending($ssl);
-			$buf .= Net::SSLeay::peek( $ssl,$pending );	 # will not block
-			if ( !$eon ) {
-				my $pos = index( $buf,"\n\n");
-				next if $pos<0; # newlines not found
-				$eon = $pos+2;	# pos after second newline
+		my $was_blocking = $self->blocking(1);
+		while ( $size>length($buf)) {
+			my $rv = $self->sysread($buf,$size-length($buf),length($buf));
+			if ( ! defined $rv ) {
+				next if $!{EINTR} or $!{EAGAIN};
+				$self->blocking(0) if ! $was_blocking;
+				return;
+			} elsif ( ! $rv ) {
+				last
 			}
-			# $eon >= 2	 == bytes incl last known \n
-			while ( index( $buf,"\n",$eon ) == $eon ) {
-				# the next char ist \n too
-				$eon++;
-			}
-			last if $eon < length($buf); # found last \n before end of buf
 		}
-		if ( $eon > 0 ) {
-			# found something
-			# readed peeked data until $eon from $ssl
-			return Net::SSLeay::ssl_read_all( $ssl,$eon );
-		} else {
-			# found nothing
-			# return all what we have
-			if ( my $l = length($buf)) {
-				return Net::SSLeay::ssl_read_all( $ssl,$l );
-			} else {
-				return $self->error( "SSL read error" );
+		$self->blocking(0) if ! $was_blocking;
+		return $buf;
+	}
+
+	my ($delim0,$delim1) = $/ eq '' ? ("\n\n","\n"):($/,'');
+
+	if ( Net::SSLeay::OPENSSL_VERSION_NUMBER() < 0x0090601f ) {
+		# no usable peek - need to read byte after byte
+		die "empty \$/ is not supported if I don't have peek" if $delim1 ne '';
+		my $buf = '';
+		my $was_blocking = $self->blocking(1);
+		while (1) {
+			my $rv = $self->sysread($buf,1,length($buf));
+			if ( ! defined $rv ) {
+				next if $!{EINTR} or $!{EAGAIN};
+				$self->blocking(0) if ! $was_blocking;
+				return;
+			} elsif ( ! $rv ) {
+				last
 			}
+			index($buf,$delim0) >= 0 and last;
+		}
+		$self->blocking(0) if ! $was_blocking;
+		return $buf;
+	}
+
+
+	# find first occurence of $delim0 followed by as much as possible $delim1
+	my $buf = '';
+	my $eod = 0;  # pointer into $buf after $delim0 $delim1*
+	my $was_blocking = $self->blocking(1);
+	my $ssl = $self->_get_ssl_object or return;
+	while (1) {
+
+		# block until we have more data or eof
+		my $poke = Net::SSLeay::peek($ssl,1);
+		if ( ! defined $poke or $poke eq '' ) {
+			next if $!{EINTR} or $!{EAGAIN};
+		}
+
+		my $skip = 0;
+
+		# peek into available data w/o reading
+		my $pending = Net::SSLeay::pending($ssl);
+		if ( $pending and 
+			( my $pb = Net::SSLeay::peek( $ssl,$pending )) ne '' ) {
+			$buf .= $pb
+		} else {
+			$self->blocking(0) if ! $was_blocking;
+			return $buf eq '' ? ():$buf;
+		};
+		if ( !$eod ) {
+			my $pos = index( $buf,$delim0 );
+			if ( $pos<0 ) {
+				$skip = $pending
+			} else {
+				$eod = $pos + length($delim0); # pos after delim0
+			}
+		}
+
+		if ( $eod ) {
+			if ( $delim1 ne '' ) {
+				# delim0 found, check for as much delim1 as possible
+				while ( index( $buf,$delim1,$eod ) == $eod ) {
+					$eod+= length($delim1);
+				}
+			}
+			$skip = $pending - ( length($buf) - $eod );
+		}
+
+		# remove data from $self which I already have in buf
+		while ( $skip>0 ) {
+			if ($self->sysread(my $p,$skip,0)) {
+				$skip -= length($p);
+				next;
+			}
+			$!{EINTR} or $!{EAGAIN} or last;
+		}
+
+		if ( $eod and ( $delim1 eq '' or $eod < length($buf))) {
+			# delim0 found and there can be no more delim1 pending
+			last
 		}
 	}
+	$self->blocking(0) if ! $was_blocking;
+	return substr($buf,0,$eod);
 }
 
 sub close {
@@ -801,6 +866,7 @@ sub close {
 
 	if ( ! $close_args->{_SSL_in_DESTROY} ) {
 		untie( *$self );
+		undef ${*$self}{_SSL_fileno};
 		return $self->SUPER::close;
 	}
 	return 1;
@@ -818,13 +884,14 @@ sub stop_SSL {
 		} else {
 			my $fast = $stop_args->{SSL_fast_shutdown};
 			my $status = Net::SSLeay::get_shutdown($ssl);
-			if ( $status == SSL_RECEIVED_SHUTDOWN
-				|| ( $status != 0 && $fast )) {
-				# shutdown done
+			if ( $fast && $status != 0) {
+				# shutdown done, either status has  
+				# SSL_SENT_SHUTDOWN or SSL_RECEIVED_SHUTDOWN or both,
+				# so the handshake is at least in process
 				$shutdown_done = 1;
-			} else {
+			} elsif ( ( $status & SSL_SENT_SHUTDOWN ) == 0 ) {
 				# need to initiate/continue shutdown
-				local $SIG{PIPE} = sub{};
+				local $SIG{PIPE} = 'IGNORE';
 				for my $try (1,2 ) {
 					my $rv = Net::SSLeay::shutdown($ssl);
 					if ( $rv < 0 ) {
@@ -838,10 +905,15 @@ sub stop_SSL {
 						$shutdown_done = 1;
 						last;
 					} else {
-						# shutdown partly finished (e.g. one direction)
+						# shutdown partly initiated (e.g. one direction)
 						# call again
 					}
 				}
+			} elsif ( $status & SSL_RECEIVED_SHUTDOWN ) {
+				# SSL_SENT_SHUTDOWN is done already (previous if-case)
+				# and because SSL_RECEIVED_SHUTDOWN is done also we
+				# consider the shutdown done
+				$shutdown_done = 1;
 			}
 		}
 
@@ -896,6 +968,13 @@ sub _get_ssl_object {
 	return $ssl;
 }
 
+# _get_ctx_object is for internal use ONLY!
+sub _get_ctx_object {
+	my $self = shift;
+	my $ctx_object = ${*$self}{_SSL_ctx};
+	return $ctx_object && $ctx_object->{context};
+}
+
 # default error for undefined arguments
 sub _invalid_object {
 	return IO::Socket::SSL->error("Undefined IO::Socket::SSL object");
@@ -1027,7 +1106,7 @@ sub dump_peer_certificate {
 		},
 		# rfc 2818
 		http => {
-			wildcards_in_cn	 => 1,
+			wildcards_in_cn	 => 'anywhere',
 			wildcards_in_alt => 'anywhere',
 			check_cn         => 'when_only',
 		},
@@ -1088,7 +1167,8 @@ sub dump_peer_certificate {
 			#  make sure that Socket6 was loaded properly
 			UNIVERSAL::can( __PACKAGE__, 'inet_pton' ) or croak
 				q[Looks like IPv6 address, make sure that Socket6 is loaded or make "use IO::Socket::SSL 'inet6'];
-			$ipn = inet_pton( $identity ) or croak "'$identity' is not IPv6, but neither IPv4 nor hostname";
+			$ipn = inet_pton(AF_INET6,$identity) 
+				or croak "'$identity' is not IPv6, but neither IPv4 nor hostname";
 		} elsif ( $identity =~m{^\d+\.\d+\.\d+\.\d+$} ) {
 			 # definitly no hostname, try IPv4
 			$ipn = inet_aton( $identity ) or croak "'$identity' is not IPv4, but neither IPv6 nor hostname";
@@ -1456,6 +1536,10 @@ sub new {
 
 	Net::SSLeay::CTX_set_verify($ctx, $verify_mode, $verify_callback);
 
+	if ( my $cb = $arg_hash->{SSL_create_ctx_callback} ) {
+		$cb->($ctx);
+	}
+
 	$ctx_object = { context => $ctx };
 	$ctx_object->{has_verifycb} = 1 if $verify_callback;
 	DEBUG(3, "new ctx $ctx" );
@@ -1470,6 +1554,7 @@ sub new {
 		$ctx_object->{session_cache} = IO::Socket::SSL::Session_Cache->new( $size );
 	}
 
+
 	return bless $ctx_object, $class;
 }
 
@@ -1805,6 +1890,20 @@ in the same call to new() will be ignored unless the context supplied was invali
 Note that, contrary to versions of IO::Socket::SSL below v0.90, a global SSL context
 will not be implicitly used unless you use the set_default_context() function.
 
+=item SSL_create_ctx_callback
+
+With this callback you can make individual settings to the context after it
+got created and the default setup was done.
+The callback will be called with the CTX object from Net::SSLeay as the single
+argument.
+
+Example for limiting the server session cache size:
+
+  SSL_create_ctx_callback => sub { 
+      my $ctx = shift;
+	  Net::SSLeay::CTX_sess_set_cache_size($ctx,128);
+  }
+
 =item SSL_session_cache_size
 
 If you make repeated connections to the same host/port and the SSL renegotiation time
@@ -1950,8 +2049,8 @@ match it checks against the common name, but there are no wildcards allowed.
 
 =item http (rfc2818), alias is www
 
-Extended wildcards in subjectAltNames are possible, e.g. *.example.org or
-even www*.example.org. Wildcards in the common name are not allowed. The common
+Extended wildcards in subjectAltNames and common name are possible, e.g. 
+*.example.org or even www*.example.org. The common
 name will be only checked if no names are given in subjectAltNames.
 
 =item smtp (rfc3207)
@@ -2185,6 +2284,10 @@ Non-blocking and timeouts (which are based on non-blocking) are not
 supported on Win32, because the underlying IO::Socket::INET does not support
 non-blocking on this platform.
 
+If you have a server and it looks like you have a memory leak you might 
+check the size of your session cache. Default for Net::SSLeay seems to be 
+20480, see the example for SSL_create_ctx_callback for how to limit it.
+
 =head1 LIMITATIONS
 
 IO::Socket::SSL uses Net::SSLeay as the shiny interface to OpenSSL, which is
